\documentclass[structure=flat, language=pt]{cweb}

\usepackage[brazil]{babel}
\usepackage[latin1]{inputenc}
\usepackage{geometry}

\geometry{
verbose,
letterpaper,
tmargin=2cm,
bmargin=2cm,
lmargin=2cm,
rmargin=2cm,
headheight=0.5cm,
headsep=0.2cm,
footskip=0cm
}

\pagestyle{myheadings}
\markright{{\sc Gerador de rvores Binrias -- Mdulo} {\tt bin-tree}}

\newcommand{\bin}[2]{\left( \begin{array}{c} #1\\#2 \end{array} \right)}

\begin{document}

@* Introduo. O mdulo {\tt bin-tree} disponibiliza vrias funes que podem
ser utilizadas com rvores binrias, as principais podendo ser utilizadas para
computar parmetros sobre rvores binrias geradas aleatriamente.

Outro conjunto de funes pode ser utilizado em conjunto com o pacote PStricks,
disponvel para \LaTeX \ ou \TeX. Estas funes podem ser utilizadas para criar
figuras das rvores binrias geradas pelo mdulo.

@ {\it Instrues.} Para que o mdulo seja utilizado por um programa, basta incluir
neste o arquivo {\tt bin-tree.h} e compilar o arquivo {\tt bin-tree.c} juntamente
com o restante do programa. O mdulo faz uso da biblioteca de funes matemticas.

Antes de poder utilizar qualquer funo disponibilizada pelo mdulo, deve-se chamar
a funo |init_bin_tree_gen| passando como argumento o maior nmero de ns que uma
rvore a ser gerada pelo mdulo vai possuir. Apenas rvores com esta quantidade de
ns ou menos podero ser geradas.

Antes de encerrar seu funcionamento, o programa principal dever chamar a funo
|deinit_bin_tree_gen| para liberar a memria alocada pelo mdulo.

Alm destas funes, o mdulo oferece outras duas funes, que so rapidamente
descritas abaixo. Para maiores instrues de uso, veja as sees adequadas.

\begin{itemize}
\item |random_tree_params|. Calcula a altura e o IPL de uma rvore binria de $N$ ns
internos gerada aleatriamente.
\item |draw_random_tree|. Gera cdigo compatvel com o pacote PStricks do \LaTeX\ ou
\TeX\ que representa uma rvore binria de $N$ ns internos gerada aleatriamente.
Tambm retorna a altura e o IPL da rvore gerada.
\end{itemize}

@(bin-tree.h@>=
@<Funes exportadas@>

@ A estrutura geral do mdulo se encontra abaixo. 

@c
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#define FALSE 0
#define TRUE !FALSE
@<Variveis globais@>@/
@<Funes privadas@>@/
@<Gerao de rvores binrias@>

@* Gerao de rvores binrias. Para que todas as rvores binrias tenham a mesma
probabilidade de serem geradas por nosso algoritmo, ele deve de alguma forma levar
em considerao a probabilidade de uma rvore binria de $N$ ns internos possuir
$L$ ns em sua subrvore esquerda e $R$ ns em sua subrvore direita.

Sabemos que a quantidade de rvores binrias com $N$ ns internos  dada pela 
seguinte expresso:

\[
C_N = \frac{1}{N + 1} \bin{2N}{N}
\]

Ento, a probabilidade de uma rvore com $N$ ns internos ter em sua subrvore
esquerda $L$ ns internos e em sua subrvore direita $R$ ns internos seria

\begin{equation}\label{eqn:prob}
P_N(L) = \frac{C_L C_R}{C_N},
\end{equation}

\noindent onde $R = N - 1 - L$.

Utilizando a aproximao de Stirling, podemos concluir que, para $L \neq 0$ e
$L \neq N-1$,

\[
P_N(L) \approx \frac{(N+1) \sqrt{\pi N}}{4 \pi (L+1)(R+1) \sqrt{LR}}.
\]

Agora, fazendo $\rho = \sqrt{\pi}/4 \pi = 0.141047396...$, teramos

\begin{equation}\label{eqn:probapprox}
P_N(L) \approx \rho \cdot \frac{(N+1) \sqrt{N}}{(L+1)(R+1) \sqrt{LR}}.
\end{equation}

Para $L = 0$ ou $R = 0$, teramos segundo a equao~(\ref{eqn:prob})

\begin{equation}\label{eqn:problzero}
P_N(0) = \frac{N+1}{4N - 2}.
\end{equation}

Para fazermos um algoritmo de gerao de rvores binrias, vamos utilizar as
equaes (\ref{eqn:probapprox}) e (\ref{eqn:problzero}) para calcularmos as
probabilidades necessrias. Manteremos estes valores em uma tabela, para evitarmos
ter de recalcul-los.

@ Nossa tabela de probabilidades vai ser um vetor de $N+1$ posies. Na $i$-sima
posio, teremos um ponteiro para um outro vetor de $i$ posies. Na $j$-sima
posio deste ltimo vetor, teremos o valor

\[
F_i(j) = \sum_{k = 0}^j P_i(k),
\]

\noindent ou seja, o valor acumulado das probabilidades de uma rvore binria com
$i$ ns internos ter uma subrvore esquerda com uma quantidade de ns internos no
intervalo $[0, j] \cap {\rm \bf Z}$.

Os valores de $P_N(L)$ so simtricos. Entretanto, para adicionarmos velocidade aos
clculos, no vamos utilizar esta propriedade para poupar espao.

Para evitar ter de calcular razes de muitos valores, manteremos tambm um vetor
com as razes de todos os nmeros inteiros entre $0$ e $N$.

@d RHO 0.141047396 /* Utilizada na gerao das probabilidades */
@<Variveis globais@>=
static double* sqr = NULL; /* |sqr[i] = | $\sqrt{i}$ */
static double** prob_table = NULL; /* Tabela de probabilidades */
static int N = 0; /* ndice da ltima posio de |prob_table| */


@ A tabela ser alocada e gerada quando o mdulo for inicializado. Para tanto,
precisaremos conhecer seu tamanho, que ser igual a |N+1|. Poderemos ento gerar
rvores aleatrias com at |N| ns internos.

A funo |init_bin_tree_gen| devolve |TRUE| se tudo correr bem e |FALSE| se houver 
algum erro.

@<Gerao de rvores...@>=
int init_bin_tree_gen(int n)
{
  if(n < 2) return FALSE;
  @<Aloca espao para |prob_table| e |sqr|@>@;
  @<Preenche |sqr|@>@;
  @<Preenche tabela de probabilidades@>@;
  free(sqr); /* No utilizaremos as razes para mais nada */
  sqr = NULL;
  N = n;
  return TRUE;
}

@ @<Aloca espao para...@>=
{
  int i;
  sqr = malloc((n+1) * sizeof(double));
  if(sqr == NULL) return FALSE;

  prob_table = malloc((n+1) * sizeof(double*)); 
  if(prob_table == NULL) return FALSE; 

  for(i = 2; i <= n; i++) { 
    prob_table[i] = malloc(i * sizeof(double)); 
    if(prob_table[i] == NULL) return FALSE; 
  }
}

@ @<Preenche |sqr|@>=
{
  int i;
  sqr[0] = 0; sqr[1] = 1;
  for(i = 2; i <= n; sqr[i++] = sqrt(i));
}

@ As primeiras duas posies de |prob_table| existem apenas para facilitar a
manipulao de ndices. No as utilizamos de qualquer outra maneira.

@<Preenche tabela de...@>=
{
  int i;
  for(i = 2; i <= n; i++) {
    int j, max = i-1;
    double rho = RHO * (i+1) * sqr[i]; /* Constante modificada para diminuir clculos */
    prob_table[i][0] = (double) (i+1) / (4*i - 2);
    for(j = 1; j < max; j++)
      prob_table[i][j] = prob_table[i][j-1] + (rho / ((j+1) * (i-j) * sqr[j] * sqr[i-1-j]));
    prob_table[i][max] = 2.0; /* Com isso, evitamos possveis problemas de aproximao */
  }
}

@ O programa cliente deve informar ao mdulo que a memria alocada para as tabelas
deve ser desalocada antes do programa ser encerrado. Isso  feito atravs da funo
|deinit_bin_tree_gen|.

@<Gerao de rvores...@>+=
void deinit_bin_tree_gen()
{
  int i;
  for(i = 2; i <= N; free(prob_table[i++]));
  free(prob_table);
  prob_table = NULL;
  N = 0;
}

@ @<Funes exportadas@>=
extern int init_bin_tree_gen(int);
extern void deinit_bin_tree_gen();

@ Agora apresentamos a funo |random_tree_paramsR|, que calcula a altura e o IPL de uma rvore
binria gerada aleatriamente. A rvore no  de fato armazenada: a prpria funo de
gerao calcula seus parmetros a medida que prossegue, fazendo isso de forma recursiva.

Esta funo no  visvel ao programa cliente, sendo acessada atravs de um envelope que
ser mostrado mais a frente.

Os argumentos recebidos so a profundidade da raiz da rvore sendo gerada e o nmero de ns
que possui. A profundidade deve ser ZERO na primeira chamada.

@<Gerao de rvores...@>+=
static void random_tree_paramsR(int n, int depth)
{
  @<Executa base da recurso@>@;
  rtree_ipl += depth; /* Atualizamos apenas o IPL, a altura atualizamos na base */
  @<Sorteia tamanho das subrvores e executa recurso@>@;
}

@ @<Executa base...@>=
if(n == 0)
  return;
else if(n == 1) {
  if(depth + 1 > rtree_height) rtree_height = depth + 1;
  rtree_ipl += depth;
  return;
}

@ @<Sorteia tamanho das subrvores e executa...@>=
{
  int i;
  double temp = (double) rand() / RAND_MAX;
  for(i = 0; i < n; i++)
    if(temp < prob_table[n][i]) break;
  random_tree_paramsR(i, depth+1); /* Subrvore esquerda */
  random_tree_paramsR(n-1-i, depth+1); /* Subrvore direita */
}

@ @<Funes privadas@>=
static void random_tree_paramsR(int, int);

@ Antes de apresentarmos o envelope para a funo |random_tree_paramsR|, 
definimos as variveis globais que assumimos em seu desenvolvimento.

@<Variveis globais@>+=
static int rtree_height = 0; /* Altura da rvore sendo gerada */
static int rtree_ipl = 0; /* IPL da rvore sendo gerada */

@ Finalmente, o evelope. As variveis apontadas por |height| e |ipl| 
contero no retorno estes parmetros.

@<Gerao de rvores...@>+=
void random_tree_params(int n, int* height, int* ipl)
{
  rtree_height = 0;
  rtree_ipl = 0;
  random_tree_paramsR(n, 0);
  *height = rtree_height;
  *ipl = rtree_ipl;
}

@ @<Funes exportadas@>+=
extern void random_tree_params(int, int*, int*);

@ Para criarmos figuras das rvores a medida que as geramos utilizamos
basicamente uma cpia da funo anterior, tambm colocando o cdigo de
gerao. Pouca coisa precisa ser mudada para tanto.

@<Gerao de rvores...@>+=
static void draw_random_treeR(int n, int depth)
{
  @<Se |n==1| ou |n==0|, cria uma folha ou n externo@>@;
  rtree_ipl += depth; /* Atualizamos apenas o IPL, a altura atualizamos na base */
  fprintf(dtree_output, "\\pstree{%s}{\n", dtree_nodetype);
  @<Sorteia tamanho das subrvores e as gera recursivamente@>@;
  fprintf(dtree_output, "\n}");
}

@ @<Sorteia tamanho das subrvores e as gera...@>=
{
  int i;
  double temp = (double) rand() / RAND_MAX;
  for(i = 0; i < n; i++)
    if(temp < prob_table[n][i]) break;
  draw_random_treeR(i, depth+1); /* Subrvore esquerda */
  draw_random_treeR(n-1-i, depth+1); /* Subrvore direita */
}

@ Os ns externos so ``impressos'' como ns nulos, que apenas ocupam espao.
Nenhum n nulo  gerado para folhas, que so explicitamente impressas.

@<Se |n==1|...@>=
if(n == 0) {
  fprintf(dtree_output, "\\Tn ");
  return;
}
else if(n == 1) {
  if(depth+1 > rtree_height) rtree_height = depth+1;
  fprintf(dtree_output, "%s ", dtree_nodetype);
  rtree_ipl += depth;
  return;
}

@ @<Funes privadas@>+=
static void draw_random_treeR(int, int);

@ Para que a funo funcione, so necessrias duas variveis globais que guardam
informaes sobre a sada desejada. Vamos defin-las agora.

@<Variveis globais@>+=
static const char* dtree_nodetype = NULL; /* Tipo do n */
static FILE* dtree_output = NULL; /* Stream para o qual ser mandada a sada */

@ O envelope da funo |draw_random_treeR|  ligeiramente diferente do envelope
da funo |random_tree_params|. Algumas configuraes da sada devem ser conhecidas
de incio e impressas junto a raiz da rvore, portanto, a raiz  processada
explicitamente.

Os argumentos |levelsep|, |treesep| e |nodetype| so obrigatrios. Eles definem a distncia 
entre dois nveis da rvore, a distncia entre duas subrvores e o tipo do n que ser
utilizado, repectivamente. O valor esperado  uma sequncia de caracteres como |"2cm"|, 
por exemplo.
No caso de |nodetype|, o argumento  algo como |"\\Tc*{2pt}"|\footnote{Veja a documentao
do pacote PStricks}, por exemplo.
O argumento |output|  um ponteiro para o {\it stream} para o qual ser jogada a sada.

@<Gerao de rvores...@>+=
void draw_random_tree(int n, int* height, int* ipl, const char* levelsep, const char* treesep,
		      const char* nodetype, FILE* output)
{
  rtree_height = 0; 
  rtree_ipl = 0;
  dtree_output = output;
  dtree_nodetype = nodetype;
  fprintf(output, "\\pstree[levelsep=%s, treesep=%s]{%s}{\n", levelsep,
	  treesep, nodetype);
  @<Sorteia tamanho das subrvores e chama a funo |draw_random_treeR|@>@;
  fprintf(output, "\n}");
  *height = rtree_height;
  *ipl = rtree_ipl;
}

@ @<Sorteia tamanho das subrvores e chama...@>=
{
  int i;
  double temp = (double) rand() / RAND_MAX;
  for(i = 0; i < n; i++)
    if(temp < prob_table[n][i]) break;
  draw_random_treeR(i, 1); /* Subrvore esquerda */
  draw_random_treeR(n-1-i, 1); /* Subrvore direita */
}

@ @<Funes exportadas@>+=
extern void draw_random_tree(int, int*, int*, const char*, const char*, const char*, FILE*);

@
\end{document}
\fi
\inx
\fin
\con
