\documentclass[structure=flat, language=pt]{cweb}

\usepackage[brazil]{babel}
\usepackage[latin1]{inputenc}
\usepackage{geometry}

\geometry{
verbose,
letterpaper,
tmargin=2cm,
bmargin=2cm,
lmargin=2cm,
rmargin=2cm,
headheight=0.5cm,
headsep=0.2cm,
footskip=0cm
}

\pagestyle{myheadings}
\markright{{\sc Gerador de rvores Binrias de Busca -- Mdulo} {\tt bs-tree}}

\newcommand{\bin}[2]{\left( \begin{array}{c} #1\\#2 \end{array} \right)}

\begin{document}

@* Introduo. O mdulo {\tt bs-tree} permite a criao de rvores binrias de
busca atravs da insero de um nmero qualquer de elementos numa ordem aleatria.
Os elementos em questo so nmeros inteiros. As rvores so criadas principalmente
para que sua altura e IPL sejam calculados.

Tambm permitimos a insero de elementos passados atravs de um vetor pelo programa
cliente, na ordem em que aparecem. Isso pode ser usado em conjunto com as funes
que geram desenhos de rvores utilizando o formato do pacote PStricks, disponvel para
\LaTeX\ ou \TeX, para gerar uma rvore especfica.

O pacote disponibiliza o seguinte conjunto de funes. Veja suas definies para
dados mais detalhados.

\begin{itemize}
\item |init_bst|. Cria uma rvore vazia e retorna um ponteiro para ela.
\item |destroy_bst|. Destri uma rvore criada com a funo acima, liberando a memria
que ocupa.
\item |get_bst_height|. Calcula a altura da rvore.
\item |get_bst_ipl|. Calcula o IPL da rvore.
\item |gen_random_bst|. Gera uma rvore binria de busca atravs da insero de uma
permutao dos elementos do vetor $a = (0, 1, \dots, N-1)$, onde $N$  o nmero de ns internos
da rvore a ser gerada.
\item |insert_array|. Insere os elementos de um vetor numa rvore de busca binria,
na ordem crescente de seus ndices.
\item |draw_bst|. Desenha uma rvore de busca binria utilizando a conveno do
pacote PStricks.
\end{itemize}

@(bs-tree.h@>=
@<Typedefs@>@/
@<Funes exportadas@>

@ A estrutura geral do mdulo  a seguinte:

@c
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#define FALSE 0
#define TRUE !FALSE

@<Estruturas de dados@>@/
@<Variveis globais@>@/
@<Funes de manipulao de rvores@>@/
@<Funes privadas@>@/
@<Funes exportadas@>@/
@<Gerao de rvores binrias de busca@>@/
@<Gerao de desenhos@>

@* Estruturas de dados. Vamos permitir que o programa cliente manipule vrias
rvores binrias de busca de tamanhos diferentes. Como o objetivo principal
do mdulo  permitir a gerao de estatsticas sobre rvores binrias de busca,
devemos tambm permitir que uma determinada rvore seja reutilizada diversas
vezes com eficincia.

@ A representao que escolhemos  a usual para rvores binrias de busca. Nesta
representao, um n interno  uma estrutura que carrega um nmero inteiro (sua
informao) e dois ponteiros para sua subrvore esquerda e direita. Um n externo
 um ponteiro nulo (|NULL|).

@<Estruturas de...@>=
typedef struct node* link;
struct node {
  int info;  /* Informao carregada pelo n, que tambm  a chave */
  link left;  /* Ponteiro para a raiz da subrvore esquerda */
  link right;  /* Ponteiro para a raiz da subrvore direita */
};

@ Uma rvore para o programa cliente no ser um simples ponteiro para um n, mas
sim uma estrutura mais complicada, que guarda mais informaes, como altura e IPL.

Para aumentarmos a eficincia na reutilizao das rvores, no vamos permitir que
ns internos sejam adicionados durante a execuo. Isso significa que o nmero de
ns internos da rvore deve ser conhecido durante sua criao. Desta forma, podemos
alocar todos os ns internos necessrios durante a inicializao, o que torna o 
programa mais eficiente.

@<Estruturas de...@>+=
typedef struct tree {
  link root; /* Raiz da rvore */
  int N; /* Nmero mximo de ns internos */
  int height; /* Altura da rvore */
  int ipl; /* IPL da rvore */
  int free_index; /* ndice do primeiro n livre em |vertex_array| */
  link vertex_array; /* Essencial para a alocao/liberao da memria */
}* bs_tree;

@ O programa cliente vai ter acesso s nossas rvores binrias atravs de um ponteiro
para |struct tree|, |bs_tree|. Este  o nico tipo exportado pelo mdulo.

@<Typedefs@>=
typedef struct tree* bs_tree;

@ Apenas rvores com um nmero mximo de ns internos maior que ZERO podem ser inicializadas.
A funo de inicializao recebe o nmero mximo de ns internos que a rvore poder
ter. Ela retorna um ponteiro para a nova rvore se tudo correr bem, |NULL| caso contrrio.

@<Funes de manipulao...@>=
bs_tree init_bst(int n)
{
  bs_tree new_tree; /* Conter o ponteiro de retorno */

  if(n <= 0) return NULL;
  @<Aloca e inicializa estrutura |tree|@>@;
  @<Aloca ns internos@>@;

  return new_tree;
}

@ @<Aloca e inicializa estrutura |tree|@>=
{
  new_tree = malloc(sizeof(struct tree));
  if(new_tree == NULL) return NULL;
  new_tree->root = NULL;
  new_tree->N = n;
  new_tree->height = new_tree->ipl = 0;
}

@ Alocaremos os ns internos em blocos, atravs da funo |calloc|, o que torna o programa
mais eficiente. Guardaremos o ponteiro retornado em |vertex_array|.

@<Aloca ns internos@>=
{
  link temp = calloc(n, sizeof(struct node));
  if(temp == NULL) return NULL;
  new_tree->vertex_array = temp;
  new_tree->free_index = 0;
}

@ A insero na rvore  feita atravs de uma simples funo recursiva. Utilizamos variveis
globais para aumentar a profundidade mxima da recurso.

Durante a insero, j atualizamos o IPL e a altura da rvore. Para isso, utilizamos o argumento
|depth|, que contm a altura da raiz sendo considerada. Inicialmente, deve ser igual a ZERO.

@<Funes de mani...@>+=
static link insertR(link root, int depth)
{
  if(root == NULL)
    @<Insere novo n e o retorna@>@;
  if(root->info > ins_num) root->left = insertR(root->left, depth+1);
  else root->right = insertR(root->right, depth+1);
  return root;
}

@ @<Insere novo...@>=
{
  link new_node = ins_bst->vertex_array + ins_bst->free_index;
  ins_bst->free_index++; /* O prximo n livre */
  new_node->info = ins_num;
  new_node->left = new_node->right = NULL;
  if(ins_bst->height <= depth) ins_bst->height = depth+1;
  ins_bst->ipl += depth;
  return new_node;
}

@ Vamos agora definir as variveis globais utilizadas pela funo |insertR|. A varivel
|ins_bst|  um ponteiro para a rvore na qual est sendo inserido o elemento, contido
na varivel |ins_num|.

@<Variveis globais@>+=
bs_tree ins_bst; /* rvore sendo considerada na insero */
int ins_num; /* Elemento sendo inserido */

@ Precisaremos tambm de um pequeno envelope para a funo |insertR|, o que facilitar
seu uso.

@<Funes de mani...@>+=
static void insert(bs_tree bst, int num)
{
  ins_bst = bst;
  ins_num = num;
  bst->root = insertR(bst->root, 0);
}

@ Com o esquema de alocao que utilizamos, liberar a memria ocupada por uma rvore
fica simples: basta liberar o vetor apontado por |vertex_array| e a estrutura |tree|
em si.

@<Funes de mani...@>+=
void destroy_bst(bs_tree bst)
{
  free(bst->vertex_array);
  free(bst);
}

@ Para que o programa cliente possa conhecer a altura e o IPL de uma rvore, disponibilizamos
duas pequenas funes.

@<Funes de mani...@>=
int get_bst_height(bs_tree bst)
{
  return bst->height;
}

int get_bst_ipl(bs_tree bst)
{
  return bst->ipl;
}

@ @<Funes exportadas@>+=
extern bs_tree init_bst(int);
extern void destroy_bst(bs_tree);
extern int get_bst_height(bs_tree);
extern int get_bst_ipl(bs_tree);

@ @<Funes privadas@>+=
static link insertR(link, int);
static void insert(bs_tree, int);

@* Gerao de rvores binrias de busca. Vamos disponibilizar funes para gerar uma rvore
binria de busca a partir de uma permutao qualquer de um determinado nmero de chaves.

Tambm permitimos que uma rvore seja gerada a partir da insero de uma determinada sequncia
de chaves {\it na ordem em que aparecem}.

@ Para gerarmos uma rvore de $N$ ns internos, vamos criar uma permutao $a'$ do vetor
$a = (0, 1, \dots, N-1)$ e inserir na rvore os elementos $a'_0, a'_1, \dots, a'_{N-1}$, 
nesta ordem.

@<Gerao de rvores...@>=
void gen_random_bst(bs_tree bst)
{
  @<Limpeza da rvore@>@;
  @<Criao e permutao do vetor $a$@>@;
  @<Insero dos elementos@>@;
}

@ Antes de podermos inserir elementos na rvore, devemos limp-la, ou seja, fazer com que 
o primeiro
n livre para insero seja o primeiro do vetor |vertex_array| e modificar o valor de 
sua altura e IPL 
para ZERO.
Graas a maneira como alocamos os ns internos, isso  bem simples.

@<Limpeza da...@>=
{
  bst->root = NULL;
  bst->height = bst->ipl = 0;
  bst->free_index = 0;
}

@ Representaremos o vetor $a$ no prprio vetor |vertex_array|, ao invs de criar outro vetor
temporrio.

@<Criao e permu...@>=
{
  int i;
  link temp = bst->vertex_array;
  for(i = 0; i < bst->N; temp[i++].info = i);
  for(i = 0; i < bst->N; i++) {
    int j = (int) ((double) bst->N * rand() / (RAND_MAX + 1.0));
    int d = temp[j].info;
    temp[j].info = temp[i].info;
    temp[i].info = d;
  }
}

@ Agora inserimos os elementos do vetor |vertex_array| na ordem |vertex_array[0]|, $\dots$,
|vertex_array[N-1]|. Note que, quando inserimos os elementos utilizando a funo |insertR|,
o primeiro elemento livre do vetor |vertex_array|  utilizado pela funo na insero. Desta forma,
no corremos o risco de perder dados.

@<Insero dos eleme...@>=
{
  int i;
  link temp = bst->vertex_array;
  for(i = 0; i < bst->N; insert(bst, temp[i++].info));
}

@ Para inserirmos uma sequncia de chaves, passada pelo programa cliente, na rvore, procedemos
de maneira semelhante ao que  feito pela funo |gen_random_bst|.

@<Gerao de rvores...@>=
void insert_array(bs_tree bst, int array_size, int* array)
{
  @<Limpeza da rvore@>@;
  @<Insero do vetor |array|@>@;
}

@ @<Insero do vetor...@>=
{
  int i;
  for(i = 0; i < array_size; insert(bst, array[i++]));
}

@ @<Funes exportadas@>+=
extern void gen_random_bst(bs_tree);
extern void insert_array(bs_tree, int, int*);

@* Gerao de desenhos. As funes que disponibilizamos para gerar desenhos de rvores binrias
geram, na verdade, cdigo compatvel com o pacote PStricks, disponvel para \LaTeX\ ou \TeX.

Vrias opes do pacote podem ser utilizadas para personalizar as rvores desenhadas. Veja a
documentao do pacote para maiores detalhes.

@ Vamos desenhar a rvore recursivamente, percorrendo-a em pr-ordem. Para aumentarmos o limite
da profundidade de recurso, vamos utilizar algumas variveis globais para controlar o processo.
Estas variveis vo conter, basicamente, os estilos dos ns a serem criados.

@<Variveis globais@>=
static int dtree_control_seq; /* Veja abaixo */
static const char* dtree_node_type; /* Estilo do n */
static FILE* dtree_output; /* {\it Stream} para o qual ser enviada a sada */

@ A funo que realmente desenha a rvore  uma funo recursiva que a percorre em pr-ordem,
imprimindo o cdigo necessrio. Esta  uma funo interna, a funo exportada  na verdade
apenas um envelope.

@d MAX_BUFFER 255
@<Gerao de desenhos@>=
static void draw_bstR(link root)
{
  static char buffer[MAX_BUFFER];
  @<Se |root| aponta para uma folha ou n externo, executa base da recurso@>@;
  @<Imprime a raiz@>@;
  draw_bstR(root->left);
  draw_bstR(root->right);
  fprintf(dtree_output, "}\n");
}

@ A base da recurso no ocorre simplesmente quando |root == NULL|, mas tambm quando 
|root != NULL| e |root->left == root->right == NULL|.

Fazemos isso pois queremos distinguir o caso em que estamos tratando de uma folha, situao
na qual no fazemos a impresso de ns externos, e o caso em que estamos tratando de um n externo,
filho de algum n interno que no  uma folha. Neste ltimo caso, devemos imprimir
o n externo, para que o {\it lay-out} da rvore seja adequado.

@<Se |root| aponta...@>=
{
  if(root == NULL) { /* Se isso ocorre, |root| no  filho de uma folha */
    fprintf(dtree_output, "\\Tn ");
    return;
  }
  else if(root->left == NULL && root->right == NULL) {
    if(dtree_control_seq) {
      sprintf(buffer, dtree_node_type, root->info);
      fprintf(dtree_output,"%s ", buffer);
    }
    else fprintf(dtree_output, "%s ", dtree_node_type);
    return;
  }
}

@ Se a base no foi executada, devemos imprimir a raiz da rvore em questo, para
depois imprimirmos a subrvore esquerda e a subrvore direita. Permitimos que o
programa cliente utilize, para espeficiar o tipo de n a ser utilizado, as sequncias
de controle permitidas pela funo |sprintf|. A varivel |dtree_control_seq| indica
se isso ocorre ou no.
    
@<Imprime a raiz@>=
{
  if(dtree_control_seq) { /* Devemos considerar sequncias de controle */
    sprintf(buffer, dtree_node_type, root->info);
    fprintf(dtree_output, "\\pstree{%s}{\n", buffer);
  }
  else fprintf(dtree_output, "\\pstree{%s}{\n", dtree_node_type);
}

@ @<Funes privadas@>+=
static void draw_bstR(link);

@ O envelope da funo |draw_bstR| faz um pouco mais do que simplesmente chamar a funo
recursiva, uma vez que algumas informaes precisam ser impressas junto a raiz da rvore
a ser impressa.

Os argumentos |level_sep| e |tree_sep| especificam a distncia entre dois nveis da rvore
e a distncia entre duas subrvores. Os valores esperados so sequncias de caracteres do
tipo |"2mm"|. 

O parmetro |node_type| permite a especificao do tipo de n interno a ser utilizado.
O valor esperado  uma sequncia do tipo |"\\Tc*{2pt}"|. Como os ns contm informao,
permitimos que esta informao seja impressa. Se o argumento |control_seq|  verdadeiro,
sequncias de controle como as utilizadas pela funo |sprintf| sero substitudas pela
informao do n.

Por exemplo, para imprimir o nmero associado a um n dentro dele no desenho, pode-se
utilizar em |node_type| a sequncia |"\\Toval{%d}"| e passar o argumento |control_seq|
como sendo verdadeiro.

Finalmente, |output| deve conter um ponteiro para o {\it stream} de sada.

@<Gerao de desenhos@>+=
void draw_bst(bs_tree bst, int control_seq, const char* node_type, const char* level_sep, 
	      const char* tree_sep, FILE* output)
{
  if(control_seq) {
    char buffer[MAX_BUFFER];
    sprintf(buffer, node_type, bst->root->info);
    fprintf(output, "\\pstree[levelsep=%s, treesep=%s]{%s}{\n", level_sep, tree_sep, buffer);
  }
  else fprintf(output, "\\pstree[levelsep=%s, treesep=%s]{%s}{\n", level_sep, tree_sep, node_type);
  dtree_control_seq = control_seq;
  dtree_node_type = node_type;
  dtree_output = output;
  draw_bstR(bst->root->left);
  draw_bstR(bst->root->right);
  fprintf(output, "}\n");
}

@ @<Funes exportadas@>+=
extern void draw_bst(bs_tree, int, const char*, const char*, const char*, FILE*);

@
\end{document}
\fi
\inx
\fin
\con
